/*  This file is part of Jellyfish.

    This work is dual-licensed under 3-Clause BSD License or GPL 3.0.
    You can choose between one of them if you use this work.

`SPDX-License-Identifier: BSD-3-Clause OR  GPL-3.0`
*/


#include <jellyfish/mer_dna.hpp>

namespace jellyfish { namespace mer_dna_ns {
const char* const error_different_k = "Length of k-mers are different";
const char* const error_short_string = "Input string is to short";
} } // namespace jellyfish { namespace mer_dna_ns
